#ifndef _TMVECTOR2_HPP
#define _TMVECTOR2_HPP
/** 
	@file vector2.hpp
	@brief Defines tempest::vector<T,2>.
	@author ototoi / Toru Matsuoka
	@date 2004/05/12 
*/

#include<cstddef>//std::size_t
#include<cmath>
#include<algorithm>
#include<stdexcept>

#include"vector.hpp" //For static polimorphizm.

namespace tempest{

template<class T, std::size_t Sz>class vector;//pre
template<class Self, class Type, std::size_t Sz>struct vector_expression;//pre
	
/** 
 *	@class vector<T,2>
 *	@brief vector class template (Specific 2-dimentional)
 *	@todo	
 *	@code
 *		tempest::vector<double,2> v;	
 *	@endcode
 *	@bug	
 */	
template<class T>class vector<T,2>;//predacal.
	
template<class T>
class vector<T,2>:public vector_base< vector<T,2>, T, 2>{
	
public:
	
	//-----------------------------------------------
	//type defines
	typedef		T									value_type;
	typedef		T&									reference;							  
	typedef		const T&							const_reference;
	typedef		vector<T,2>							this_type;

	typedef		std::size_t							size_type;
	typedef		std::ptrdiff_t						difference_type;
	
	typedef		T*									iterator;
	typedef		const T*							const_iterator;
	
#ifdef _TEMPEST_USE_META_
	typedef		typename param_type<T>::type			param_type;
#else
	typedef		T										param_type;
#endif
	
	
public:
	static const size_type c_size = 2;//container size	
	
public:
	//-----------------------------------------------
	//functions for iterator
	iterator		begin()			{return element;}
	iterator		end()			{return element+c_size;}
	const_iterator	begin()	const	{return element;}
	const_iterator	end()	const	{return element+c_size;}
	
#define SC_(a) static_cast<T>(a)	
	//-----------------------------------------------
	//constructors and destructor
	vector(){}
	
	vector(param_type _x, param_type _y):m0(_x),m1(_y){}
	
	vector(const this_type &rhs):m0(rhs.m0),m1(rhs.m1){}
	
	template<class X>
	explicit vector(const vector<X,2> &rhs):m0(SC_(rhs[0])),m1(SC_(rhs[1])){}
	
	template<class X>
	explicit vector(const X * rhs):m0(SC_(rhs[0])),m1(SC_(rhs[1])){}
	
	template<class Self,class X>
	explicit vector(const vector_base<Self,X,2> & rhs):
		m0(SC_(static_cast<const Self &>(rhs)[0])),
		m1(SC_(static_cast<const Self &>(rhs)[1])){}
	
	~vector(){}
	//
	//void swap(this_type& rhs){std::swap(*this,rhs);}
	
	
	//-----------------------------------------------
	//inserters
	this_type& operator= (const this_type &rhs){
		m0 = rhs.m0;
		m1 = rhs.m1;
		return *this;
	}

	template<class X>
	this_type& operator= (const vector<X,2> &rhs){
		m0 = SC_(rhs[0]);
		m1 = SC_(rhs[1]);
		return *this;
	}
	
	template<class Self,class X>
	this_type& operator= (const vector_expression<Self,X,2> & rhs){	
		m0 = SC_(static_cast<const Self &>(rhs)[0]);
    	m1 = SC_(static_cast<const Self &>(rhs)[1]);
    	return *this;
	}
	
	template<class Self,class X>
	this_type& operator= (const vector_base<Self,X,2> & rhs){
		m0 = SC_(static_cast<const Self &>(rhs)[0]);
    	m1 = SC_(static_cast<const Self &>(rhs)[1]);
		return *this;
	}
	
	template<class IT>
	void assign( IT start, IT end ){
		assert(std::difference(start,end)<=c_size);//debug
		
		std::copy(start,end,begin());				
	}
	
	void assign( size_type num, param_type val ){
		std::fill_n(begin(),(num<c_size)?num:c_size,val);		
	}

#undef SC_
	
	//-----------------------------------------------
	//capacity
	size_type size ()	  const { return c_size; }
	size_type max_size () const { return c_size; }
	bool	  empty ()	  const { return false;	 }
	
	//-----------------------------------------------
	//operators
	
#define DECLARE_OP_EQUAL( OP )							\
	this_type& operator OP (const this_type & rhs){		\
		m0 OP rhs.m0;									\
		m1 OP rhs.m1;									\
		return *this;									\
	}													\
	template<class X>									\
	this_type& operator OP (const vector<X,2> & rhs){	\
		m0 OP rhs[0];									\
		m1 OP rhs[1];									\
		return *this;									\
	}
	
	DECLARE_OP_EQUAL( += )
	DECLARE_OP_EQUAL( -= )
	DECLARE_OP_EQUAL( *= )
	DECLARE_OP_EQUAL( /= )

#undef DECLARE_OP_EQUAL
	
	//--------------------------------
	this_type& operator*= (param_type rhs){
		m0 *= rhs;
		m1 *= rhs;
	
		return *this;
	}	 
	
	this_type& operator/= (param_type rhs){
		m0 /= rhs;
		m1 /= rhs;
	
		return *this;
	}	 
	
	template<class X>
	this_type& operator*= (const X& rhs){
		m0 *= rhs;
		m1 *= rhs;
	
		return *this;
	}	 
	template<class X>
	this_type& operator/= (const X& rhs){
		m0 /= rhs;
		m1 /= rhs;
	
		return *this;
	}	 
	//--------------------------------
	
	
	T& operator[](size_type i){
		return element[i];
	}
	
	param_type operator[](size_type i) const {
		return element[i];
	}
	
	T& at(size_type i){
		if(c_size<=i){throw std::out_of_range("tempest::vector2");}
		return element[i];
	}
	const T& at(size_type i) const {
		if(c_size<=i){throw std::out_of_range("tempest::vector2");}
		return element[i];
	}
	
	
	//-----------------------------------------------
	//utilities
	value_type length() const {
		return std::sqrt( 
			static_cast<double>( sqr_length() ) 
		);
	}
	
	value_type sqr_length() const {
		return (
			(m0 * m0) + (m1 * m1) 
		);
	}
	this_type& normalize(){
		T length = sqr_length();		//||V||^2
		if (length == T(0)) return *this;
	
		length = RSQRT( length  );	// 1 / ||V||	
		
		m0 *= length;
		m1 *= length;

		return *this;
	}
	
	
	
	//-----------------------------------------------
	//debug code
public:
	const char* debug()const{return "tempest::vector<T,2>";}
	
private:
	/** @union */
	union{
		/** @struct */
		struct{
			T m0,m1;
		};
		T element[2];
	};

};//End of class.


//-----------------------------------------------
//Not a member!
//-----------------------------------------------
	
//-----------------------------------------------
//operators 
	
	//See vector.hpp!

//-----------------------------------------------	
//utility functions

	//normalize
	//length
	//sqr_length

template<class T>
inline T dot(const vector<T,2> &lhs,const vector<T,2> &rhs){
	return (lhs[0] * rhs[0]) + (lhs[1] * rhs[1]) ;
}

//-----------------------------------------------
//compare
template<class T>
inline bool operator== (const vector<T,2> &lhs, const vector<T,2> &rhs){
	return (lhs[0] == rhs[0])&&(lhs[1] == rhs[1]);
}

	
//-----------------------------------------------
//
}//End of namespace.

	
#include "vector_functions.hpp"




#endif
